/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.streams.jackson;

import org.apache.streams.data.util.RFC3339Utils;

import com.fasterxml.jackson.core.JsonParser;
import com.fasterxml.jackson.databind.DeserializationContext;
import com.fasterxml.jackson.databind.deser.std.StdDeserializer;
import org.joda.time.DateTime;
import org.joda.time.format.DateTimeFormat;
import org.joda.time.format.DateTimeFormatter;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

/**
 * StreamsDateTimeDeserializer is a supporting class for
 * @see {@link org.apache.streams.jackson.StreamsJacksonMapper}
 *
 * <p></p>
 * Converting date-time strings other than RFC3339 to joda DateTime objects requires
 * additional formats to be provided when instantiating StreamsJacksonMapper.
 */
public class StreamsDateTimeDeserializer extends StdDeserializer<DateTime> implements Serializable {

  private List<DateTimeFormatter> formatters = new ArrayList<>();

  private static final Logger LOGGER = LoggerFactory.getLogger(StreamsDateTimeDeserializer.class);

  protected StreamsDateTimeDeserializer(Class<DateTime> dateTimeClass) {
    super(dateTimeClass);
  }

  protected StreamsDateTimeDeserializer(Class<DateTime> dateTimeClass, List<String> formats) {
    super(dateTimeClass);
    for ( String format : formats ) {
      try {
        formatters.add(DateTimeFormat.forPattern(format));
      } catch (Exception ex) {
        LOGGER.warn("Exception parsing format " + format);
      }
    }
  }

  /**
   * Applies each additional format in turn, until it can provide a non-null DateTime
   */
  @Override
  public DateTime deserialize(JsonParser jpar, DeserializationContext context) throws IOException {

    DateTime result = RFC3339Utils.parseToUTC(jpar.getValueAsString());
    Iterator<DateTimeFormatter> iterator = formatters.iterator();
    while ( result == null && iterator.hasNext()) {
      DateTimeFormatter formatter = iterator.next();
      result = formatter.parseDateTime(jpar.getValueAsString());
    }
    return result;
  }
}
